c has_cubic.f, (Aug 12, 1995)
c ----------------------------------------------------------------------
      PROGRAM has_cubic
c########################################################################
c A driver program for estimating function f by HAS procedure,
c     using cubic spline basis functions, from data:
c     y_i = f(x_i) + noise, i = 1,2,...,n                                
c     where x_i is in [0,1].
c                                                                        
c Subroutines called:                                                    
c     presele_fast, nbgcv, predcu,
c     fnull, rk, (both given in basiscu.f),
c     suppl.f (a group of subroutines from BLAS and LINPACK),            
c     and gcvpack.f (a group of subroutines from GCVPACK).
c                                                                        
c########################################################################
      INTEGER NOBS, MAXBAS
      PARAMETER(NOBS=256,MAXBAS=50)
c########################################################################
c The simpliest way to use this program is to replace just above two     
c     parameters, where                                                  
c                                                                        
c     NOBS:   number of observations.                                    
c     MAXBAS: maximum number of basis functions to be used in HAS fit. 
c                                                                        
c !Caution: For NOBS=512, MAXBAS=200, it takes about 30 seconds          
c     on a DEC Alpha; for NOBS=2048, MAXBAS=100, about 10 minutes!       
c     Also a lot of memory!                                              
c########################################################################
      INTEGER NB,NPAR,NNULL,MAXTBL
      DOUBLE PRECISION X(NOBS),Y(NOBS),IDF,COEF(MAXBAS),
     & HASFIT(NOBS),HASDEV,HASDF
      PARAMETER(NNULL=2,MAXTBL=100)
c########################################################################
c     NNULL:  dimension of null space.
c     MAXTBL: maximum number of evenly spaced values to be used
c             in the initial grid search for lambda, the smoothing
c             parameter in the penalized regression step of HAS.
c             Used in GCVPACK/dsnsm.
c             NTBL set before calling dsnsm has to be less or
c             equal to MAXTBL.
c Inputs:
c     X:      covariate.
c     Y:      response.
c     IDF:    the inflated degrees of freedom for each adaptively 
c             selected basis function. (1.2 is recommended for general 
c             purpose, it should be, however, between 1 and 2).
c     NPAR:   number of basis functions to be used in penalized 
c             regression. (for now, it is recommended as equal to NB).
c 
c Outputs:
c     NB:     number of basis functions chosen by modified GCV criterion.
c     COEF:   coefficients of the penalized regression on the 
c             selected basis functions. (first two are for the two
c             basis functions in null space, i.e., 1 and (x-.5).
c     HASFIT: HAS estimates of the noiseless response at original
c             data points.
c     HASDEV: estimate of standard deviation of noise from final HAS fit
c             (only for reference).
c     HASDF:  degrees of freedom from final HAS fit (only for reference).
c
c Variables used only in dsnsm (GCVPACK):
c     MAXTBL,LWA,LIWA,NTBL,IOUT,IWORK,DES,SIGMA,ADIAG,TAU,LAMLIM,
c     DOUT,SVALS,TBL,AUXTBL,WORK.
c     The meanings of them can be found in dsnsm.f in GCVPACK.
c########################################################################
      INTEGER NBAS, LWA, LIWA
      PARAMETER(NBAS = NNULL+NOBS, 
     & LIWA = 2*MAXBAS-NNULL,
     & LWA = (MAXBAS-NNULL)*(MAXBAS-2*NNULL+2+NOBS)+MAXBAS+NOBS )
      INTEGER ISB(MAXBAS),IWK(NBAS),INFO,
     & NTBL,IOUT(3),IWORK(LIWA)
      DOUBLE PRECISION BAS(NOBS,NBAS),YY(NOBS),RSS(MAXBAS),
     & WKA(NOBS,MAXBAS),GCV(MAXBAS),DUM,DELTA,U(NOBS),V(NOBS),
     & DES(NOBS,MAXBAS),SIGMA(MAXBAS,MAXBAS),ADIAG(NOBS),TAU,LAMLIM(2),
     & DOUT(5),SVALS(NOBS),TBL(MAXTBL,3),AUXTBL(3,3),WORK(LWA)
      DOUBLE PRECISION FNULL, RK, PREDCU

      WRITE(*,*) "NOBS = ", NOBS
      WRITE(*,*) "MAXBAS = ", MAXBAS

c input data (data file has to have two columns, x and y)
      OPEN(UNIT=20,FILE="samplecu.data",STATUS="UNKNOWN")
      DO 25 I =1,NOBS
         READ(20,*) X(I),Y(I)
 25   CONTINUE
      CLOSE(UNIT=20)

c compute all the candidate basis functions
      DO 31 M = 1,NNULL
         DO 30 J=1,NOBS
            BAS(J,M) = FNULL (M,X(J))
 30      CONTINUE
 31   CONTINUE
      DO 33 J=1,NOBS
         DO 32 I=1,NOBS
            BAS(I,J+NNULL) = RK (X(I), X(J))
 32      CONTINUE
 33   CONTINUE

c call presele to select basis functions one by one (i.e. order them).
      CALL DCOPY(NOBS,Y(1),1,YY(1),1)
c     BAS and YY will be destroyed after calling presele.
      CALL PRESELE(BAS,NOBS,NBAS,YY,NNULL,MAXBAS,ISB,RSS,IWK,WKA)
      OPEN(UNIT=22,FILE="out.isb",STATUS="UNKNOWN")
      DO 34 I = 1,MAXBAS
         WRITE(22,*) ISB(I)
 34   CONTINUE
      CLOSE(22)
      OPEN(UNIT=24,FILE="out.rss",STATUS="UNKNOWN")
      DO 35 I = 1,MAXBAS
         WRITE(24,*) RSS(I)
 35   CONTINUE
      CLOSE(24)

c call nbgcv to decide the number of basis functions to be used
      WRITE(*,*) "Enter IDF (in general, 1.2 is recommended, but any"
      WRITE(*,*) "           number between 1 and 2 may be tried)"
      READ(*,*) IDF 
      CALL NBGCV(NOBS,NNULL,RSS,MAXBAS,IDF,NB,GCV)
c
      WRITE(*,*) "The number of basis functions chosen by GCV"
      WRITE(*,*) "with IDF ", SNGL(IDF)," is ", NB, " (NB)."
      WRITE(*,*) "Enter NPAR (< MAXBAS, for now, NB is recommended)"
      READ(*,*) NPAR
c call dsnsm from GCVPACK to compute the penalized regression problem
c   compute the first NB selected basis
      DO 41 M = 1,NNULL
         DO 40 I = 1,NOBS
            DES(I,M) = FNULL(M,X(I))
 40      CONTINUE
 41   CONTINUE

c if NPAR = NNULL, only an ordinary regression is needed.
      IF(NPAR.EQ.NNULL) GOTO 1000
      DO 43 I = 1,NOBS
         DO 42 J = NNULL+1,NPAR
            DES(I,J) = RK(X(I),X(ISB(J)-NNULL))
 42      CONTINUE
 43   CONTINUE
c   compute SIGMA.
c   Sometimes a matrix DELTA*IDENTITY is added to SIGMA to ensure 
c   the solution of the linear system from dsnsm is numerically stable,
c   otherwise dsnsm may give the error message 221.
      DELTA = .001D0
      DO 45 I=1,NPAR
         CALL DSET(NPAR,0.D0,SIGMA(1,I),1)
 45   CONTINUE
      DO 47 I=NNULL+1,NPAR
         DO 46 J=NNULL+1,NPAR
            SIGMA(I,J) = RK(X(ISB(I)-NNULL),X(ISB(J)-NNULL))
 46      CONTINUE
         SIGMA(I,I) = SIGMA(I,I) + DELTA
 47   CONTINUE
c   
      CALL DCOPY(NOBS,Y,1,HASFIT,1)
      TAU = 10.D0
      NTBL = 100
      JOB = 1000
      CALL DSNSM(DES,NOBS,HASFIT,SIGMA,MAXBAS,NOBS,NPAR,NNULL,ADIAG,
     &     TAU,LAMLIM,NTBL,DOUT,IOUT,COEF,SVALS,TBL,MAXTBL,AUXTBL,
     &     IWORK,LIWA,WORK,LWA,JOB,INFO)
      WRITE(*,*) "INFO (from GCVPACK/dsnsm) = ", INFO
      IF(INFO .gt. 0) STOP "GCVPACK/dsnsm error"
      OPEN(UNIT=26,FILE="out.coef",STATUS="UNKNOWN")
      DO 48 I = 1,NPAR
         WRITE(26,*) COEF(I)
 48   CONTINUE
      CLOSE(26)

      HASDEV = SQRT(DOUT(3)/(DBLE(NOBS)-(DBLE(NOBS)-DOUT(4)-
     & DBLE(NNULL))*IDF-DBLE(NNULL)))
      HASDF = DBLE(NOBS)-DOUT(4)
      GOTO 2000
 1000 CONTINUE
      CALL DQRDC(DES,NOBS,NOBS,NNULL,U,IWK,V,0)
      CALL DQRSL(DES,NOBS,NOBS,NNULL,U,Y,DUM,V,
     & COEF,DUM,HASFIT,00101,INFO)
      HASDEV = SQRT(RSS(NNULL)/(DBLE(NOBS)-DBLE(NNULL)))
      HASDF = NNULL
 2000 CONTINUE
c output and summary of HAS estimate
      WRITE(*,*) "   NPAR,   HASDEV,   HASDF"
      WRITE(*,100) NPAR,HASDEV,HASDF
 100      FORMAT(1X,I5,1X,E12.4,1X,F7.1)
      OPEN(UNIT = 30, FILE = "out.hasfit",STATUS="UNKNOWN")
      DO 50 I = 1,NOBS
         WRITE(30,*) HASFIT(I), PREDCU(X(I),NNULL,NPAR,X,ISB,COEF)
 50   CONTINUE
      CLOSE(30)
c fitted values both  directly from dsnsm and from pred using COEF are 
c reported in order to compare them. If there are significant differences,
c then COEF's maybe haven't been computed stably in dsnsm. One solution is
c to increase DELTA (say by 10 times), and refit.
      STOP
      END
