c presele.f, (Aug 14 1995, 23:25)
c ----------------------------------------------------------------------
      SUBROUTINE PRESELE(BAS,N,NBAS,Y,NNULL,MAXBAS,ISB,RSS,IWK,WKA)
c########################################################################c
c A subroutine for PREliminary SELEcting basis functions.
c     Select, from NBAS candidates, MAXBAS basis functions
c     which have the largest RSS reduction by LS fit in turn.
c
c Use QR decomposition done by a sequence of Householder reflections
c     to compute each RSS in LS fit (without pivoting).
c     (ref. SEBER (1977) pp 338-340 )
c     QR decomposition and the application of it are updated;
c     BAS and Y are destroyed after calling.
c
c Subroutines used: 
c     those in suppl.f: DCOPY, DSET, ISET, (indirect: DAXPY),
c     those in the end of this file: HOUSEHOLDER, APPLYHOUSEHOLDER 
c Functions used: DDOT, DNRM2 (they are in suppl.f too)
c
c Input:
c     N:      the number of observation
c     NBAS:   the number of candidate basis functions
c     BAS:    (N,NBAS), the basis functions evaluated at the
c             observation points
c     Y:      (N), the response
c     NNULL:  the number of basis functions which will be forced
c             into the seected subset
c     MAXBAS: the maximum number of basis functions to be chosen
c     IWK:    (NBAS), work vector,
c             IWK(I) is the index of whether the I-th basis function
c             in BAS has been chosen or not;  0 -- no, 1 -- yes
c     WKA:    (N,MAXBAS), work array 
c Output:
c     ISB:    (MAXBAS), the column index of chosen basis functions
c             in BAS
c     RSS:    (MAXBAS), RSS(I) is the residual sum of squares of the LS 
c             fit when regressing Y on the first I chosen basis
c             functions.
c
c April 3, 1995, 0:30 by Zhen Luo
c########################################################################c
      INTEGER N, NBAS, NNULL, MAXBAS, ISB(*), IWK(*)
      DOUBLE PRECISION BAS(N,*), Y(*), RSS(*), WKA(N,*), PREV, CURR
      INTEGER I, J, IB, JB
      DOUBLE PRECISION DDOT
      CALL ISET(NBAS,0,IWK(1),1)
c     first NNULL basis functions in BAS are forced into selected set.
      ISB(1) = 1
      IWK(1) = 1
      CALL HOUSEHOLDER(N,1,BAS(1,1),WKA(1,1))
      DO 10 IB = 2,NNULL
         CALL APPLYHOUSEHOLDER(N,IB-1,WKA(1,IB-1),Y)
         DO 9 J = IB,NBAS
            CALL APPLYHOUSEHOLDER(N,IB-1,WKA(1,IB-1),BAS(1,J))
 9       CONTINUE
         ISB(IB) = IB
         IWK(IB) = 1
         CALL HOUSEHOLDER(N,IB,BAS(1,IB),WKA(1,IB))
 10   CONTINUE
c
      DO 20 IB = NNULL+1, MAXBAS
         CALL APPLYHOUSEHOLDER(N,IB-1,WKA(1,IB-1),Y)
         RSS(IB-1) = DDOT(N-IB+1,Y(IB),1,Y(IB),1)
         ISB(IB) = 0
         PREV = 0.D0
         CURR = 0.D0
         DO 18 JB = NNULL+1,NBAS
            IF(IWK(JB).EQ.1) GOTO 18
            CALL APPLYHOUSEHOLDER(N,IB-1,WKA(1,IB-1),BAS(1,JB))
            CURR = (DDOT(N-IB+1,BAS(IB,JB),1,Y(IB),1))**2/
     &           DDOT(N-IB+1,BAS(IB,JB),1,BAS(IB,JB),1)
            IF(CURR.LE.PREV) GOTO 18
            PREV = CURR
            ISB(IB) = JB
 18      CONTINUE
         IWK(ISB(IB)) = 1
         CALL HOUSEHOLDER(N,IB,BAS(1,ISB(IB)),WKA(1,IB))
 20   CONTINUE
      CALL APPLYHOUSEHOLDER(N,MAXBAS,WKA(1,MAXBAS),Y)
      RSS(MAXBAS) = DDOT(N-MAXBAS,Y(MAXBAS+1),1,Y(MAXBAS+1),1)
      RETURN
      END 
      SUBROUTINE HOUSEHOLDER(N,J,A,W)
c     Compute the Householder matrix for eliminating the last
c     (N-J) elements of N-vector A. Store the transformation matrix
c     information in N-vector W (the first (J-1) elements of W
c     are zeros). The Householder matrix H = I - 2*W*W'.
      INTEGER N,J, II
      DOUBLE PRECISION A(*), W(*), S, DNRM2
      S = DNRM2(N-J+1,A(J),1)
      CALL DSET(J-1,0.D0,W(1),1)
      W(J) = A(J) + SIGN(S,A(J))
      S = SQRT(2.D0*S*(S+ABS(A(J))))
      W(J) = W(J)/S
      DO 1 II = J+1,N
         W(II) = A(II)/S
 1    CONTINUE
      RETURN
      END
      SUBROUTINE APPLYHOUSEHOLDER(N,J,W,V)
c     Apply a Householder reflection, i.e. multiply a Householder
c     matrix H = I - 2*W*W', to the N-vector V, where W is a N-vector
c     and the first (J-1) elements of it are zeros.
      INTEGER N,J
      DOUBLE PRECISION W(*), V(*), T, DDOT
      T = - DDOT(N-J+1,W(J),1,V(J),1) * 2.D0
      CALL DAXPY(N-J+1,T,W(J),1,V(J),1)
      RETURN
      END

